import { AST_NODE_TYPES, TSESTree } from '@typescript-eslint/utils';
import createRule from '../util/createRule';

const memoizeComponent = createRule({
    name: 'memoize-component',
    meta: {
        docs: {
            description: 'React components should be memoized',
            recommended: 'warn',
        },
        messages: {
            general: 'Memoize this component',
        },
        type: 'problem',
        schema: [
            {
                type: 'object',
                properties: {},
                additionalProperties: false,
            },
        ],
        fixable: 'code',
    },
    defaultOptions: [],
    create(context) {
        const verifyVariableDeclaration = (
            node: TSESTree.VariableDeclaration
        ) => {
            const declaration = node.declarations[0];

            if (declaration.type === AST_NODE_TYPES.VariableDeclarator) {
                const init = declaration.init;
                if (!init) {
                    // Not a component
                    return;
                }

                const name =
                    declaration.id.type === AST_NODE_TYPES.Identifier
                        ? declaration.id.name
                        : undefined;

                if (
                    !name ||
                    // Shallow checking for component declaration
                    name.charAt(0) !== name.charAt(0).toUpperCase()
                ) {
                    return;
                }

                const defaultExportNode = (
                    (node.parent as any).body as TSESTree.BaseNode[]
                )?.find?.(item => {
                    return (
                        item.type === AST_NODE_TYPES.ExportDefaultDeclaration
                    );
                }) as TSESTree.ExportDefaultDeclaration | undefined;

                if (init.type === AST_NODE_TYPES.ArrowFunctionExpression) {
                    if (!defaultExportNode) {
                        // TODO: Only report if component
                        context.report({
                            node,
                            messageId: 'general',
                            *fix(fixer) {
                                yield fixer.insertTextBefore(init, 'memo(');

                                yield fixer.insertTextAfter(init, ')');
                            },
                        });

                        return;
                    }

                    const isDefaultExportMemoized = (() => {
                        const exportDeclaration = (
                            defaultExportNode as TSESTree.ExportDefaultDeclaration
                        ).declaration;

                        if (
                            exportDeclaration &&
                            exportDeclaration.type ===
                                AST_NODE_TYPES.CallExpression
                        ) {
                            const callee = exportDeclaration.callee;
                            if (
                                callee.type === AST_NODE_TYPES.Identifier &&
                                callee.name === 'memo'
                            ) {
                                return true;
                            }
                        }
                    })();

                    if (isDefaultExportMemoized) {
                        return;
                    }

                    context.report({
                        node: defaultExportNode,
                        messageId: 'general',
                        *fix(fixer) {
                            yield fixer.insertTextBefore(
                                defaultExportNode.declaration,
                                'memo('
                            );

                            yield fixer.insertTextAfter(
                                defaultExportNode.declaration,
                                ')'
                            );
                        },
                    });

                    return;
                }

                if (init.type === AST_NODE_TYPES.CallExpression) {
                    const callee = init.callee;
                    const args = init.arguments;
                    const calleeChain = args
                        .filter(
                            item =>
                                item.type === AST_NODE_TYPES.CallExpression &&
                                item.callee.type === AST_NODE_TYPES.Identifier
                        )
                        .map(
                            (item: TSESTree.CallExpressionArgument) =>
                                (
                                    (item as TSESTree.CallExpression)
                                        .callee as TSESTree.Identifier
                                ).name
                        );

                    if (callee.type === AST_NODE_TYPES.Identifier) {
                        calleeChain.push(callee.name);
                    }

                    if (!calleeChain.includes('memo')) {
                        context.report({
                            node,
                            messageId: 'general',
                            *fix(fixer) {
                                yield fixer.insertTextBefore(init, 'memo(');

                                yield fixer.insertTextAfter(init, ')');
                            },
                        });

                        return;
                    }
                }
            }
        };

        const verifyExportNamedDeclaration = (
            node: TSESTree.ExportNamedDeclaration
        ) => {
            if (node.declaration?.type === AST_NODE_TYPES.VariableDeclaration) {
                verifyVariableDeclaration(node.declaration);
            }
        };

        const findRootNode = (
            currentNode: TSESTree.Node
        ): TSESTree.Program | undefined => {
            if (currentNode.type === 'Program') {
                return currentNode;
            }

            const parent = currentNode.parent;
            if (!parent) {
                return undefined;
            }

            return findRootNode(parent);
        };

        const processDeclarations = (
            declarations: (
                | TSESTree.VariableDeclaration
                | TSESTree.ExportNamedDeclaration
            )[]
        ) => {
            for (const declaration of declarations) {
                if (declaration.type === AST_NODE_TYPES.VariableDeclaration) {
                    verifyVariableDeclaration(declaration);
                    continue;
                }

                if (
                    declaration.type === AST_NODE_TYPES.ExportNamedDeclaration
                ) {
                    verifyExportNamedDeclaration(declaration);
                }
            }
        };

        return {
            JSXElement(node) {
                if (node.parent?.type !== AST_NODE_TYPES.ReturnStatement) {
                    return;
                }

                const root = findRootNode(node);

                const declarations = (root?.body.filter(item =>
                    [
                        AST_NODE_TYPES.VariableDeclaration,
                        AST_NODE_TYPES.ExportNamedDeclaration,
                    ].includes(item.type)
                ) ?? []) as (
                    | TSESTree.VariableDeclaration
                    | TSESTree.ExportNamedDeclaration
                )[];

                processDeclarations(declarations);
            },
            JSXFragment(node) {
                if (node.parent?.type !== AST_NODE_TYPES.ReturnStatement) {
                    return;
                }

                const root = findRootNode(node);

                const declarations = (root?.body.filter(item =>
                    [
                        AST_NODE_TYPES.VariableDeclaration,
                        AST_NODE_TYPES.ExportNamedDeclaration,
                    ].includes(item.type)
                ) ?? []) as (
                    | TSESTree.VariableDeclaration
                    | TSESTree.ExportNamedDeclaration
                )[];

                processDeclarations(declarations);
            },
        };
    },
});

export default memoizeComponent;
